@@ -37,6 +37,8 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <limits.h>
+#include <math.h>
 #include <poll.h>
 
 #if HAVE_UNISTD_H
@@ -88,6 +90,8 @@
 #endif
 #include <pcap.h>
 
+#include "arping.h"
+
 #ifndef ETH_ALEN
 #define ETH_ALEN 6
 #endif
@@ -141,6 +145,7 @@ static int beep = 0;                 /* beep when reply is received. -a */
 static int reverse_beep = 0;         /* beep when expected reply absent. -e */
 static int alsototal = 0;            /* print sent as well as received. -u */
 static int addr_must_be_same = 0;    /* -A */
+static int unsolicited = 0;          /* -U */
 
 static int finddup = 0;              /* finddup mode. -d */
 static int dupfound = 0;             /* set to 1 if dup found */
@@ -150,6 +155,11 @@ static unsigned int numsent = 0;     /* packets sent */
 static unsigned int numrecvd = 0;    /* packets received */
 static unsigned int numdots = 0;     /* dots that should be printed */
 
+static double stats_min_time = -1;
+static double stats_max_time = -1;
+static double stats_total_time = 0;
+static double stats_total_sq_time = 0;
+
 /* RAWRAW is RAW|RRAW */
 static enum { NORMAL,      /* normal output */
               QUIET,       /* No output. -q */
@@ -167,6 +177,41 @@ int verbose = 0;  /* Increase with -v */
 /* Doesn't really need to be volatile, but doesn't hurt. */
 static volatile sig_atomic_t time_to_die = 0;
 
+/**
+ * Some stupid OSs (Solaris) think it's a good idea to put network
+ * devices in /dev and then play musical chairs with them.
+ *
+ * Since libpcap doesn't seem to have a workaround for that, here's arpings
+ * workaround.
+ *
+ * E.g. if the network interface is called net0, pcap will fail because it
+ * fails to open /dev/net, because it's a directory.
+ */
+static pcap_t*
+do_pcap_open_live(const char *device, int snaplen,
+                  int promisc, int to_ms, char *errbuf)
+{
+        pcap_t* ret;
+        char buf[PATH_MAX];
+
+        if ((ret = pcap_open_live(device, snaplen, promisc, to_ms, errbuf))) {
+                return ret;
+        }
+
+        snprintf(buf, sizeof(buf), "/dev/%s", device);
+        if ((ret = pcap_open_live(buf, snaplen, promisc, to_ms, errbuf))) {
+                return ret;
+        }
+
+        snprintf(buf, sizeof(buf), "/dev/net/%s", device);
+        if ((ret = pcap_open_live(buf, snaplen, promisc, to_ms, errbuf))) {
+                return ret;
+        }
+
+        /* Call original again to reset the error message. */
+        return pcap_open_live(device, snaplen, promisc, to_ms, errbuf);
+}
+
 /**
  *
  */
@@ -296,6 +341,7 @@ extended_usage()
 	       "           \"own\" the MAC address you are using.\n"
 	       "    -u     Show index=received/sent instead  of  just  index=received  when\n"
 	       "           pinging MACs.\n"
+	       "    -U     Send unsolicited ARP.\n"
 	       "    -v     Verbose output. Use twice for more messages.\n"
 	       "    -w     Time to wait between pings, in microseconds.\n");
         printf("Report bugs to: thomas@habets.se\n"
@@ -311,7 +357,7 @@ standard_usage()
 {
 	printf("ARPing %s, by Thomas Habets <thomas@habets.se>\n",
 	       version);
-        printf("usage: arping [ -0aAbdDeFpqrRuv ] [ -w <us> ] "
+        printf("usage: arping [ -0aAbdDeFpqrRuUv ] [ -w <us> ] "
                "[ -S <host/ip> ]\n"
                "              "
                "[ -T <host/ip ] "
@@ -402,6 +448,31 @@ static int get_mac_addr(const char *in,
 	return 0;
 }
 
+/**
+ *
+ */
+static void
+update_stats(double sample)
+{
+        if (stats_min_time < 0 || sample < stats_min_time) {
+                stats_min_time = sample;
+        }
+        if (sample > stats_max_time) {
+                stats_max_time = sample;
+        }
+        stats_total_time += sample;
+        stats_total_sq_time += sample * sample;
+}
+
+/**
+ *
+ */
+static double
+timespec2dbl(const struct timespec *tv)
+{
+        return tv->tv_sec + (double)tv->tv_nsec / 1000000000;
+}
+
 /**
  * max size of buffer is intsize + 1 + intsize + 4 = 70 bytes or so
  *
@@ -415,8 +486,8 @@ static char *ts2str(const struct timespec *tv, const struct timespec *tv2,
 	double f,f2;
 	int exp = 0;
 
-	f = tv->tv_sec + (double)tv->tv_nsec / 1000000000;
-	f2 = tv2->tv_sec + (double)tv2->tv_nsec / 1000000000;
+        f = timespec2dbl(tv);
+        f2 = timespec2dbl(tv2);
 	f = (f2 - f) * 1000000000;
 	while (f > 1000) {
 		exp += 3;
@@ -530,7 +601,7 @@ pingip_send()
 					  ARPOP_REQUEST,
 					  srcmac,
 					  (uint8_t*)&srcip,
-                                          (uint8_t*)ethnull,
+					  unsolicited ? (uint8_t*)ethxmas : (uint8_t*)ethnull,
 					  (uint8_t*)&dstip,
 					  NULL,
 					  0,
@@ -600,6 +671,8 @@ pingip_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet)
 			return;
 		}
 		if (dstip == ip) {
+                        update_stats(timespec2dbl(&arrival)
+                                     - timespec2dbl(&lastpacketsent));
 			switch(display) {
 			case DOT:
 				numdots++;
@@ -623,7 +696,8 @@ pingip_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet)
 				printf(" time=%s",
                                        ts2str(&lastpacketsent,
                                               &arrival,buf));
-				break; }
+                                break;
+                        }
 			case QUIET:
 				break;
 			case RAWRAW:
@@ -704,6 +778,8 @@ pingmac_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet)
 				return;
 			}
 		}
+                update_stats(timespec2dbl(&arrival)
+                             - timespec2dbl(&lastpacketsent));
 		switch(display) {
 		case QUIET:
 			break;
@@ -771,6 +847,7 @@ ping_recv(pcap_t *pcap, uint32_t packetwait, pcap_handler func)
        struct timespec endtime;
        char done = 0;
        int fd;
+       int old_received;
 
        if (verbose > 3) {
                printf("arping: receiving packets...\n");
@@ -782,6 +859,7 @@ ping_recv(pcap_t *pcap, uint32_t packetwait, pcap_handler func)
        fixup_timespec(&endtime);
 
        fd = pcap_get_selectable_fd(pcap);
+       old_received = numrecvd;
 
        for (;!done;) {
 	       int trydispatch = 0;
@@ -811,7 +889,7 @@ ping_recv(pcap_t *pcap, uint32_t packetwait, pcap_handler func)
                 * it here */
                if ((ts.tv_sec > packetwait / 1000000)
                    || ((ts.tv_sec == packetwait / 1000000)
-                       && (ts.tv_nsec > packetwait % 1000000))) {
+                       && (ts.tv_nsec/1000 > packetwait % 1000000))) {
 		       ts.tv_sec = 0;
 		       ts.tv_nsec = 1;
                        done = 1;
@@ -836,6 +914,11 @@ ping_recv(pcap_t *pcap, uint32_t packetwait, pcap_handler func)
                        r = select(fd + 1, &fds, NULL, NULL, &tv);
 		       switch (r) {
 		       case 0: /* timeout */
+                               if (display == NORMAL) {
+                                       if (numrecvd == old_received) {
+                                               printf("Timeout\n");
+                                       }
+                               }
 			       done = 1;
 			       break;
 		       case -1: /* error */
@@ -843,7 +926,7 @@ ping_recv(pcap_t *pcap, uint32_t packetwait, pcap_handler func)
 				       done = 1;
 				       sigint(0);
 				       fprintf(stderr,
-					       "arping: poll() failed: %s\n",
+					       "arping: select() failed: %s\n",
 					       strerror(errno));
 			       }
 			       break;
@@ -855,9 +938,9 @@ ping_recv(pcap_t *pcap, uint32_t packetwait, pcap_handler func)
 
 	       if (trydispatch) {
 		       int ret;
-		       if (1 != (ret = pcap_dispatch(pcap, 1,
-						     func,
-						     NULL))) {
+                       if (0 > (ret = pcap_dispatch(pcap, -1,
+                                                    func,
+                                                    NULL))) {
 			       /* rest, so we don't take 100% CPU... mostly
                                   hmm... does usleep() exist everywhere? */
 			       usleep(1);
@@ -865,7 +948,7 @@ ping_recv(pcap_t *pcap, uint32_t packetwait, pcap_handler func)
 			       /* weird is normal on bsd :) */
 			       if (verbose > 3) {
 				       fprintf(stderr,
-					       "arping: poll says ok, but "
+					       "arping: select says ok, but "
 					       "pcap_dispatch=%d!\n",
 					       ret);
 			       }
@@ -907,7 +990,7 @@ int main(int argc, char **argv)
 	dstip = 0xffffffff;
 	memcpy(dstmac, ethxmas, ETH_ALEN);
 
-	while (EOF!=(c=getopt(argc,argv,"0aAbBc:dDeFhi:I:pqrRs:S:t:T:uvw:"))) {
+	while (EOF!=(c=getopt(argc,argv,"0aAbBc:dDeFhi:I:pqrRs:S:t:T:uUvw:"))) {
 		switch(c) {
 		case '0':
 			srcip = 0;
@@ -1036,6 +1119,14 @@ int main(int argc, char **argv)
 		case 'u':
 			alsototal = 1;
 			break;
+		case 'U':
+			if (mode == PINGMAC) {
+				fprintf(stderr, "arping: -U can only be used "
+					"in IP ping mode\n");
+				exit(1);
+			}
+			unsolicited = 1;
+			break;
 		case 'v':
 			verbose++;
 			break;
@@ -1053,7 +1144,7 @@ int main(int argc, char **argv)
                 clock_getres(CLOCK_MONOTONIC, &ts);
                 printf("clock_getres() = %ld %ld\n",
                        (long)ts.tv_sec, (long)ts.tv_nsec);
-#else 
+#else
                 printf("Using gettimeofday() for time measurements\n");
 #endif
         }
@@ -1205,7 +1296,7 @@ int main(int argc, char **argv)
 	/*
 	 * pcap init
 	 */
-	if (!(pcap = pcap_open_live((char*)ifname, 100, promisc, 10, ebuf))) {
+        if (!(pcap = do_pcap_open_live(ifname, 100, promisc, 10, ebuf))) {
 		fprintf(stderr, "arping: pcap_open_live(): %s\n",ebuf);
 		exit(1);
 	}
@@ -1214,7 +1305,7 @@ int main(int argc, char **argv)
 		exit(1);
 	}
 	if (verbose > 1) {
-		printf("pcap_get_selectable(): %d\n",
+		printf("pcap_get_selectable_fd(): %d\n",
 		       pcap_get_selectable_fd(pcap));
 	}
 
@@ -1331,7 +1422,18 @@ int main(int argc, char **argv)
                        "unanswered (%d extra)\n",
                        target,numsent,numrecvd,
                        (succ < 0.0) ? 0.0 : succ,
-                       (succ < 0.0) ? (numrecvd - numsent): 0); 
+                       (succ < 0.0) ? (numrecvd - numsent): 0);
+                if (numrecvd) {
+                        double avg = stats_total_time / numrecvd;
+                        printf("rtt min/avg/max/std-dev = "
+                               "%.3f/%.3f/%.3f/%.3f ms",
+                               1000*stats_min_time,
+                               1000*avg,
+                               1000*stats_max_time,
+                               1000*sqrt(stats_total_sq_time/numrecvd
+                                         -avg*avg));
+                }
+                printf("\n");
 	}
 
         if (finddup) {
