@@ -25,8 +25,8 @@
 #include <inttypes.h>
 #endif
 
-extern uint32_t srcip,dstip;
 extern int verbose;
+void do_signal_init();
 void do_libnet_init(const char *ifname);
 void sigint(int);
 const char *arping_lookupdev_default(uint32_t srcip, uint32_t dstip,
